from dataclasses import dataclass
import time
from io import BytesIO, StringIO
from xml.dom.minidom import Element, parse
import os
from history4feed.app.settings import history4feed_server_settings as settings
import requests
from dateutil.parser import parse as parse_date
from readability import Document as ReadabilityDocument
import brotli
from types import SimpleNamespace
from .import logger
from .xml_utils import getAtomLink, getFirstChildByTag, getFirstElementByTag, getText
from .exceptions import history4feedException, UnknownFeedtypeException, FetchRedirect, ScrapflyError
import fake_useragent
from urllib.parse import urljoin

def fetch_page_with_retries(url, retry_count=3, sleep_seconds=settings.WAYBACK_SLEEP_SECONDS, **kwargs):
    ua = fake_useragent.UserAgent()
    session = requests.Session()
    session.max_redirects = 3
    headers = kwargs.get('headers', {})
    headers.update({
        "User-Agent": ua.random,
    })
    kwargs.update(headers=headers)
    error = None
    for i in range(retry_count+1):
        try:
            if i > 0:
                time.sleep(sleep_seconds * 1.5 ** (i-1))
            return fetch_page(session, url, **kwargs)
        except FatalError:
            raise
        except BaseException as e:
            error = e
            print(error)
    raise ConnectionError(f"could not fetch page after {retry_count} retries") from error
    
class FatalError(Exception):
    pass

def fetch_page(session, url, headers=None) -> tuple[bytes, str, str]:
    proxy_apikey = settings.SCRAPFLY_APIKEY
    headers = headers or {}

    if proxy_apikey:
        headers, result = fetch_with_scapfly(session, url, headers, proxy_apikey)
        return result.content.encode(), result.content_type, result.url

    logger.info(f"Fetching `{url}`")
    resp: requests.Response  = session.get(url, headers=headers)
    content = resp.content
    if not resp.ok:
        raise history4feedException(f"GET Request failed for `{url}`, status: {resp.status_code}, reason: {resp.reason}")

    # some times, wayback returns br encoding, try decompressing
    try:
        content = brotli.decompress(content)
    except Exception as err:
        logger.print(f"brotli decompress fail: {err}")
    return content, resp.headers.get("content-type"), resp.url

def fetch_with_scapfly(session, url, headers, proxy_apikey):
    logger.info(f"Fetching `{url}` via scrapfly.io")
    headers = dict((f"headers[{k}]", v) for k, v in headers.items())
    resp = session.get("https://api.scrapfly.io/scrape", params=dict(**headers, key=proxy_apikey, url=url, country="us,ca,mx,gb,fr,de,au,at,be,hr,cz,dk,ee,fi,ie,se,es,pt,nl"))
    json_data = resp.json()
    if resp.status_code != 200:
        raise ScrapflyError(json_data)
    result = SimpleNamespace(**json_data['result'])
    if result.status_code > 499:
        raise FatalError(f"Got server error {result.status_code}, stopping")
    if result.status_code > 399:
        raise history4feedException(f"PROXY_GET Request failed for `{url}`, status: {result.status_code}, reason: {result.status}")
    elif result.status_code > 299:
        raise FetchRedirect(f"PROXY_GET for `{url}` redirected, status: {result.status_code}, reason: {result.status}")
    return headers,result

def parse_feed_from_url(url):
    data, content_type, url = fetch_page_with_retries(url, retry_count=0)
    return parse_feed_from_content(data, url)

def get_full_text(link):
    try:
        page, content_type, url = fetch_page_with_retries(link)
        doc  = ReadabilityDocument(page.decode(), url=url)
        return doc.summary(), content_type
    except BaseException as e:
        raise history4feedException(f"Error processing fulltext: {e}") from e


@dataclass
class PostDict:
    link: str
    title: str
    pubdate: str
    author: str = None
    categories: list[str] = None
    description: str = "EMPTY BODY"
    content_type: str =  "text/html"

def parse_feed_from_content(data: bytes, url: str):
    feed_data = {}
    try:
        if isinstance(data, str):
            document = parse(StringIO(data))
        else:
            document = parse(BytesIO(data))
        # check if it's atom or rss
        if rss := getFirstElementByTag(document, "rss"):
            channel = getFirstElementByTag(rss, "channel")
            feed_data['description'] = getText(getFirstElementByTag(channel, "description"))
            feed_data['title'] = getText(getFirstElementByTag(channel, "title"))
            # feed_data['rel'] = getText(getFirstElementByTag(channel, "link"))

            feed_data["feed_type"] = "rss"
        elif feed := getFirstElementByTag(document, "feed"):
            feed_data['description'] = getText(getFirstElementByTag(feed, "subtitle"))
            feed_data['title'] = getText(getFirstElementByTag(feed, "title"))
            # feed_data['rel'] = getAtomLink(feed)

            feed_data["feed_type"] = "atom"
        else:
            raise UnknownFeedtypeException("feed is neither RSS or ATOM")
        feed_data["url"] = url
        return feed_data
    except BaseException as e:
        raise UnknownFeedtypeException(f"Failed to parse feed from `{url}`") from e

def get_publish_date(item):
    published = getFirstElementByTag(item, "published")
    if not published:
        published = getFirstElementByTag(item, "pubDate")
    return parse_date(getText(published))

def get_categories(entry: Element) -> list[str]:
    categories = []
    for category in entry.getElementsByTagName('category'):
        cat = category.getAttribute('term') or getText(category)
        if not cat:
            cat = category
        categories.append(cat)
    return categories

def get_author(item):
    author = getFirstElementByTag(item, "dc:creator")
    if not author:
        author = getFirstElementByTag(item, "author")
        author = getFirstElementByTag(author, "name") or author
    return getText(author)


def parse_items(elem, link):
    return PostDict(
        # element = elem,
        link = link,
        title = getText(getFirstElementByTag(elem, "title")),
        pubdate = get_publish_date(elem),
        author = get_author(elem),
        categories = get_categories(elem),
        description="",
        content_type="plain/text",
    )

def parse_posts_from_rss_feed(base_url, data) -> dict[str, PostDict]:
    entries = {}
    document = parse(BytesIO(data))
    channel = getFirstElementByTag(document, "channel")

    for item in channel.getElementsByTagName("item"):
        link = urljoin(base_url, getText(getFirstElementByTag(item, "link")).strip())
        entries[link] = parse_items(item, link)
        entries[link].description = parse_rss_description(item)
    return entries

def parse_posts_from_atom_feed(base_url, data):
    entries = {}
    document = parse(BytesIO(data))

    for item in document.getElementsByTagName("entry"):
        link = urljoin(base_url, getAtomLink(item, rel='alternate'))
        entries[link] = parse_items(item, link)
        entries[link].description, content_type = parse_atom_description(item)
        if content_type:
            entries[link].content_type = content_type
    return entries

def parse_atom_description(item: Element):
    description = ""
    if summary := getFirstChildByTag(item, "summary"):
        description = getText(summary)
    if content := getFirstChildByTag(item, "content"):
        description = getText(content)
    return description, None

def parse_rss_description(item: Element):
    return getText(getFirstChildByTag(item, "description"))



# Generated by Django 5.1.6 on 2025-03-28 10:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('history4feed', '0002_feed_freshness_alter_feed_feed_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feed',
            name='description',
            field=models.CharField(default=None, help_text='found in the <channel> of RSS output. Is always kept up to date with the latest feed import values for this property.', max_length=10240, null=True),
        ),
    ]



import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime, timedelta, timezone
from history4feed.h4fscripts.sitemap_helpers import (
    fetch_posts_links_with_serper,
    SearchIndexError,
)


@pytest.fixture
def mock_response():
    def _make_response(organic_items, credits=1):
        response = MagicMock()
        response.ok = True
        response.json.return_value = {"organic": organic_items, "credits": credits}
        print(response.json.return_value)
        return response

    return _make_response


@patch("history4feed.h4fscripts.sitemap_helpers.requests.Session.get")
def test_fetch_posts_links_basic(mock_get, mock_response):
    from_time = datetime(2024, 1, 1, tzinfo=timezone.utc)
    to_time = datetime(2024, 1, 2, tzinfo=timezone.utc)

    mock_get.return_value = mock_response(
        [
            {
                "title": "Example Title",
                "link": "https://example.com/post",
                "date": "Jan 1, 2024",
            }
        ]
    )

    result = fetch_posts_links_with_serper("example.com", from_time, to_time)

    assert len(result) == 1
    post = result["https://example.com/post"]
    assert post.title == "Example Title"
    assert post.link == "https://example.com/post"
    assert post.pubdate.date() == datetime(2024, 1, 1).date()


@patch("history4feed.h4fscripts.sitemap_helpers.requests.Session.get")
def test_fetch_posts_links_multiple_pages(mock_get, mock_response):
    mock_get.side_effect = [
        mock_response(
            [
                {"title": "Post 1", "link": "https://x1.com/1", "date": "Jan 1, 2024"},
            ]
        ),
        mock_response(
            [
                {"title": "Post 2", "link": "https://x2.com/2", "date": "Jan 5, 2024"},
            ],
            credits=19,
        ),
        mock_response([], credits=1),  # Triggers page exit
    ]

    from_time = datetime(2024, 1, 1, tzinfo=timezone.utc)
    to_time = datetime(2024, 1, 3, tzinfo=timezone.utc)

    result = fetch_posts_links_with_serper("x.com", from_time, to_time, delta_days=3)

    assert len(result) == 2
    assert ["https://x1.com/1", "https://x2.com/2"] == list(result)


@patch("history4feed.h4fscripts.sitemap_helpers.requests.Session.get")
def test_fetch_posts_handles_no_date(mock_get, mock_response):
    # Entry with no 'date'
    mock_get.return_value = mock_response(
        [{"title": "No Date", "link": "https://example.com/nodate"}]
    )

    from_time = datetime(2024, 1, 1, tzinfo=timezone.utc)
    to_time = datetime(2024, 1, 2, tzinfo=timezone.utc)

    result = fetch_posts_links_with_serper("example.com", from_time, to_time)

    assert "https://example.com/nodate" in result
    post = result["https://example.com/nodate"]
    assert post.pubdate <= to_time


@patch("history4feed.h4fscripts.sitemap_helpers.requests.Session.get")
def test_fetch_posts_raises_on_error_status(mock_get):
    mock_get.return_value.ok = False
    mock_get.return_value.status_code = 403
    mock_get.return_value.text = "Forbidden"

    from_time = datetime(2024, 1, 1, tzinfo=timezone.utc)

    with pytest.raises(SearchIndexError) as exc:
        fetch_posts_links_with_serper("fail.com", from_time)

    assert "403" in str(exc.value)



from .xml_utils import createRSSHeader, createCDataElement, createTextElement
from ..app.models import Feed, Post
from django.db.models.manager import BaseManager
from xml.dom.minidom import Document


def build_rss(feed_obj: Feed, posts_set: BaseManager[Post]):
    document, channel = createRSSHeader(feed_obj.title,  feed_obj.description, feed_obj.url, feed_obj.latest_item_pubdate)
    for post in posts_set:
        channel.appendChild(build_entry_element(post, document))

    return document.toprettyxml()

def build_entry_element(post: Post, d: Document):
    element = d.createElement('item')
    element.appendChild(createTextElement(d, "title", post.title))

    link = createTextElement(d, "link", post.link)
    link.setAttribute("href", post.link)
    element.appendChild(link)
    element.appendChild(createTextElement(d, "pubDate", post.pubdate.isoformat()))
    if post.description:
        description = post.description
        description = description
        element.appendChild(createTextElement(d, "description", description))

    for category in post.categories.all():
        element.appendChild(createTextElement(d, "category", category.name))

    if post.author:
        author = d.createElement('author')
        author.appendChild(createTextElement(d, "name", post.author))
        element.appendChild(author)
    return element




from unittest.mock import patch
import uuid
import pytest
from history4feed.app import models
from history4feed.app.models import JobState
from history4feed.app.settings import history4feed_server_settings
from rest_framework.validators import ValidationError


def test_stix_id():
    url = 'https://goo.gl'
    with patch('uuid.uuid5') as mock_uuid5:
        gen_id = models.stix_id(url)
        mock_uuid5.assert_called_once_with(uuid.UUID(str(history4feed_server_settings.HISTORY4FEED_NAMESPACE)), url)
        assert gen_id == mock_uuid5.return_value
    assert models.stix_id(url) == models.stix_id(url), "should be deterministic"

def test_normalize_url():
    assert models.normalize_url('https://example.com/net/') == 'https://example.com/net/'
    assert models.normalize_url('https://example.com//net//') == 'https://example.com/net/'
    assert models.normalize_url('http://example.com///net//a////') == 'http://example.com/net/a/'
    assert models.normalize_url('http://example.com///net') == 'http://example.com/net'
    assert models.normalize_url('http://example.com') == 'http://example.com/', "should add trailing slash when path is empty"

    with pytest.raises(ValidationError):
        models.normalize_url('kjasjkaskj')

@pytest.mark.django_db
def test_save_feed():

    orig_url = 'http://example.com///net//a////'
    feed = models.Feed.objects.create(title='my title', description='my description', url=orig_url)
    feed.save()

    assert feed.get_pretty_url() == orig_url
    assert feed.id == models.stix_id(feed.url)

def test_feed_get_pretty_url():
    orig_url = 'http://example.com'
    feed = models.Feed(title='my title', description='my description', url=orig_url)
    assert feed.get_pretty_url() == feed.url
    feed.pretty_url = 'https://example.net/pretty-url/'
    assert feed.get_pretty_url() == feed.pretty_url


@pytest.mark.django_db
@pytest.mark.parametrize(
    "original_state, new_state, expected_state, has_completion_time",
    [
        (JobState.PENDING, JobState.CANCELLED, JobState.CANCELLED, True),
        (JobState.PENDING, JobState.FAILED, JobState.FAILED, True),
        (JobState.PENDING, JobState.SUCCESS, JobState.SUCCESS, True),
        (JobState.RUNNING, JobState.SUCCESS, JobState.SUCCESS, True),
        (JobState.SUCCESS, JobState.FAILED, JobState.SUCCESS, False),
        (JobState.CANCELLED, JobState.SUCCESS, JobState.CANCELLED, False),
        (JobState.FAILED, JobState.SUCCESS, JobState.FAILED, False),
    ]
)
def test_update_state_behavior(jobs, original_state, new_state, expected_state, has_completion_time):
    job = jobs[0]
    job.state = original_state
    job.save()
    
    result = job.update_state(new_state)
    job.refresh_from_db()

    assert job.state == expected_state
    assert result == expected_state

    if has_completion_time:
        assert job.completion_time is not None
    else:
        assert job.completion_time is None


import pytest
from rest_framework import status
from history4feed.app.models import Feed, Job
from unittest.mock import MagicMock, patch

from history4feed.app.serializers import JobSerializer
from history4feed.app.views import feed_post_view

import pytest
from rest_framework import status
from history4feed.app.models import Feed, Job
from unittest.mock import patch



# from .openapi_params import FEED_PARAMS, POST_PARAMS

from history4feed.app.models import Feed, Job
from rest_framework import (
    status,
)

from tests.utils import Transport


@pytest.mark.django_db
def test_create_post_in_feed_success(client, api_schema):
    feed = Feed.objects.create(url="https://example.com/rss.xml", title="Test Feed")
    job = Job.objects.create(feed=feed)

    with (
        patch.object(feed_post_view, "new_create_post_job") as mock_new_create_post_job,
        patch(
            "history4feed.app.views.JobSerializer", wraps=JobSerializer
        ) as mock_job_s_class,
    ):
        mock_new_create_post_job.return_value = job
        url = f"/api/v1/feeds/{feed.id}/posts/"

        payload = {
            "link": "https://example.com/test-post",
            "title": "Test Post",
            "pubdate": "2024-08-20T10:00:00.000Z",
        }

        resp = client.post(url, data=payload, content_type="application/json")
        assert mock_new_create_post_job.call_args[0][1] == feed.id
        assert mock_new_create_post_job.call_args[0][0].data == payload
        assert resp.status_code == status.HTTP_201_CREATED
        mock_job_s_class.assert_called_once_with(job)

        assert resp.data["id"] == str(job.id)
        assert resp.data["feed_id"] == str(job.feed.id)

        api_schema['/api/v1/feeds/{feed_id}/posts/']['POST'].validate_response(Transport.get_st_response(resp))

@pytest.mark.django_db
def test_new_create_post_job(feed):
    posts_data = [
        {
            "link": "https://example.com/test-post",
            "title": "Test Post",
            "pubdate": "2024-08-20T10:00:00.000Z",
        }
    ]
    payload = dict(posts=posts_data)
    request = MagicMock()
    request.data = payload
    with patch(
        "history4feed.app.views.task_helper.new_patch_posts_job"
    ) as mock_new_job:
        job = feed_post_view().new_create_post_job(request, feed.id)
        post = feed.posts.first()
        mock_new_job.assert_called_once_with(feed, [post])
        assert job == mock_new_job.return_value


@pytest.mark.django_db
def test_reindex_feed(client, feed_posts, api_schema):
    feed, posts = feed_posts
    job = Job.objects.create(feed=feed)
    with (
        patch("history4feed.app.views.task_helper.new_patch_posts_job") as mock_new_job,
        patch(
            "history4feed.app.views.JobSerializer",
            wraps=JobSerializer,
        ) as mock_job_s_class,
    ):
        mock_new_job.return_value = job
        resp = client.patch(f"/api/v1/feeds/{feed.id}/posts/reindex/")
        mock_new_job.assert_called_once_with(feed, posts)
        mock_job_s_class.assert_called_once_with(job)
        assert resp.data["id"] == str(job.id)
        assert resp.data["feed_id"] == str(job.feed.id)
        api_schema['/api/v1/feeds/{feed_id}/posts/reindex/']['PATCH'].validate_response(Transport.get_st_response(resp))

